#include <winsock2.h>
#include <windows.h>
#include <process.h>

#include <time.h>
#include "dsp.h"
#include "liboddcast.h"
#include <mad.h>
#include "transcode.h"
#include "frontend.h"
#include "dsp_oddcast_wa2_bf.h"
#include "resource.h"


transcodeGlobals	globals;

int CALLBACK BigWindowHandler(HWND hwndDlg,UINT uMsg,WPARAM wParam,LPARAM lParam);
int CALLBACK EditMetadataHandler(HWND hwndDlg,UINT uMsg,WPARAM wParam,LPARAM lParam);
unsigned int oddcastThread = 0;

#ifdef WIN32
#define PopupMessage MessageBox
#else
#define PopupMessage Alert
#endif

// Forward Funtion Declarations
winampDSPModule *getModule(int which);


int initoddcast(struct winampDSPModule *this_mod);
void config(struct winampDSPModule *this_mod);
void quitoddcast(struct winampDSPModule *this_mod);
int encode_samples(struct winampDSPModule *this_mod, short int *samples, int numsamples, int bps, int nch, int srate);
// These are needed by Winamp to be a DSP module...
// best not to mess with this unless you are winamp-experienced

HWND		ghwnd_winamp;
HWND		oddcastHWND = 0;
int			timerId = 0;
int			reconnectTimerId = 0;

winampDSPModule mod =
{
	"oddcast DSP (blind friendly)",
	NULL,	// hwndParent
	NULL,	// hDllInstance
	config,
	initoddcast,
	encode_samples,
	quitoddcast
};
// Module header, includes version, description, and address of the module retriever function
winampDSPHeader hdr = { DSP_HDRVER, "oddcast DSP v2 (blind friendly)", getModule };

#ifdef __cplusplus
extern "C" {
#endif
// this is the only exported symbol. returns our main header.
__declspec( dllexport ) winampDSPHeader *winampDSPGetHeader2()
{
	return &hdr;
}
#ifdef __cplusplus
}
#endif


// getmodule routine from the main header. Returns NULL if an invalid module was requested,
// otherwise returns either mod1 or mod2 depending on 'which'.
winampDSPModule *getModule(int which)
{
	switch (which)
	{
		case 0: return &mod;
		default:return NULL;
	}
}


void inputMetadataCallback(void *pValue) {
	SetDlgItemText(oddcastHWND, IDC_METADATA, (char *)pValue);
// pValue is a string
}
void outputStatusCallback(void *pValue) {
	SetDlgItemText(oddcastHWND, IDC_STATUS, (char *)pValue);
}
void writeBytesCallback(void *pValue) {
	// pValue is a long
	static	long	startTime = 0;	
	static	long	endTime = 0;	
	static	long	bytesWrittenInterval = 0;
	char	kBPSstr[25] = "";

	long	bytesWritten = (long)pValue;

	if (bytesWritten == -1) {
		strcpy(kBPSstr, "");
		SetDlgItemText(oddcastHWND, IDC_KBPS_SENT, (char *)kBPSstr);
		startTime = 0;
		return;
	}
	if (startTime == 0) {
		startTime = time(&startTime);
		bytesWrittenInterval = 0;
	}
	bytesWrittenInterval += bytesWritten;
	globals.totalBytesWritten += bytesWritten;
	endTime = time(&endTime);
	if ((endTime - startTime) > 4) {
		int bytespersec = bytesWrittenInterval/(endTime - startTime);
		long kBPS = (bytespersec * 8)/1000;
		sprintf(kBPSstr, "%ld Kbps", kBPS);
		SetDlgItemText(oddcastHWND, IDC_KBPS_SENT, (char *)kBPSstr);
		startTime = 0;
	}
}
void outputServerNameCallback(void *pValue) {
	SetDlgItemText(oddcastHWND, IDC_SERVER_DESC, (char *)pValue);
}
void outputBitrateCallback(void *pValue) {
	SetDlgItemText(oddcastHWND, IDC_BITRATE, (char *)pValue);

}
void outputStreamURLCallback(void *pValue) {
	SetDlgItemText(oddcastHWND, IDC_DESTINATION_LOCATION, (char *)pValue);
}

void VUCallback(int leftMax, int rightMax) {
	;
}


int transcode_init(int argc, char **argv)
{
	char	*configFile = NULL;
	char	*defaultConfigFile = "transcoder.cfg";
	int	printConfig = 0;
	
	memset(&globals, '\000', sizeof(globals));

	// Setup Callbacks for all the interaction
	globals.inputMetadataCallback = inputMetadataCallback;
	globals.outputStatusCallback = outputStatusCallback;
	globals.writeBytesCallback = writeBytesCallback;
	globals.outputServerNameCallback = outputServerNameCallback;
	globals.outputBitrateCallback = outputBitrateCallback;
	globals.outputStreamURLCallback = outputStreamURLCallback;

	setServerStatusCallback(globals.outputStatusCallback);
	setGeneralStatusCallback(NULL);
	setWriteBytesCallback(globals.writeBytesCallback);
	setBitrateCallback(globals.outputBitrateCallback);
	setServerNameCallback(globals.outputServerNameCallback);
	setDestURLCallback(globals.outputStreamURLCallback);

	if (configFile == NULL) {
		configFile = defaultConfigFile;
	}
//	setConfigFileName(configFile);
	readConfigFile();
	if (printConfig) {
		printConfigFileValues();
	}

	setFrontEndType(FRONT_END_ODDCAST_PLUGIN);
	return 1;
}
// configuration. Passed this_mod, as a "this" parameter. Allows you to make one configuration
// function that shares code for all your modules (you don't HAVE to use it though, you can make
// config1(), config2(), etc...)
void config(struct winampDSPModule *this_mod)
{
	int a = 1;
	
}

void setUIMetadata(char *pMetadata)
{
	if (getLockedMetadataFlag()) {
		if (setCurrentSongTitle((char *)getLockedMetadata())) {
			if (globals.inputMetadataCallback) {
				globals.inputMetadataCallback((void *)getLockedMetadata());
			}
		}
	}
	else {
		if (setCurrentSongTitle((char *)pMetadata)) {
			if (globals.inputMetadataCallback) {
				globals.inputMetadataCallback((void *)pMetadata);
			}
		}
	}
}
void stopOddcast()
{
	disconnectFromServer();
	setForceStop(1);
	if (getLiveRecordingSetFlag()) {
		stopRecording();
	}
//	setUIMetadata("");

}
int startOddcast() 
{

	setForceStop(0);
	if (connectToServer()) {
		//UpdateWhatsPlaying();
	}
	else {
		return(0);
	}
	if (getLiveRecordingSetFlag()) {
		if (!initLiveRecording()) {
			MessageBox(oddcastHWND,"Fail to open input device", "Message", MB_OK);
			disconnectFromServer();
			return 0;
		}
		else {
			char	buf[255] = "";
			sprintf(buf, "%dHz/Stereo PCM", getLiveInSamplerate());
			setSourceDescription(buf);
			setUIMetadata(getServerDesc());
			startRecording();
		}
	}
	return 1;
}
extern "C" {
int startOddcastThread(void *)
{
	startOddcast();
	_endthread();
	return (1);
}
}

int reconnectServer()
{
	stopOddcast();
	return startOddcast();
}

VOID CALLBACK ReconnectTimer(HWND hwnd,UINT uMsg,UINT idEvent,DWORD dwTime) {
	if (!getIsConnected()) {
		if (getReconnectTrigger()) {
			if (reconnectServer()) {
				setReconnectTrigger(0);
			}
		}
	}
}

VOID CALLBACK getCurrentSongTitle(HWND hwnd,UINT uMsg,UINT idEvent,DWORD dwTime)
{
	char	songTitle[1024] = "";
	char	songTitle2[1024] = "";
	static	char currentTitle[1024] = "";
	int		windowTitle = 1;
	
	memset(songTitle2, '\000', sizeof(songTitle2));
	if (ghwnd_winamp) {
		if (GetWindowText(ghwnd_winamp, songTitle, sizeof(songTitle)) > 0) {
			if (!strncmp(songTitle, "MuchFX", strlen("MuchFX"))) {
				windowTitle = 0;
				int pos=SendMessage(ghwnd_winamp,WM_WA_IPC,0,IPC_GETLISTPOS);
				char *name = (char *)SendMessage(ghwnd_winamp,WM_WA_IPC,pos,IPC_GETPLAYLISTTITLE);
				if (name) {
					strcpy(songTitle, name);
				}
			}
			if (!windowTitle) {
				setUIMetadata(songTitle);
			}
			else {
				if (strcmp(currentTitle, songTitle)) {
					strcpy(currentTitle, songTitle);
					char	*p1 = strrchr(songTitle, '-');
					if (p1) {
						char *p2 = strchr(songTitle, '.');
						if (p2) {
							p2++;
							if (*p2 == ' ') {
								p2++;
							}
							char *p3 = strstr(p2, "- Winamp");
							if (p3) {
								*p3 = '\000';
							}
							setUIMetadata(p2);
						}
						else {
							setUIMetadata(songTitle);
						}
					}
				}
			}
		}
	}
//	SetTimer(hwnd, 1, 1000, (TIMERPROC)getCurrentSongTitle);
}
extern "C" {
void startOddcastDialog(void* arg) {
	oddcastHWND = CreateDialog(mod.hDllInstance,MAKEINTRESOURCE(IDD_ODDCAST), (HWND)arg, BigWindowHandler);
	ShowWindow(oddcastHWND, SW_SHOW);
}
}
// Here is the entry point for the Plugin..this gets called first.
int initoddcast(struct winampDSPModule *this_mod)
{
	char filename[512],*p;
	char	directory[1024] = "";
	char currentDir[1024] = "";
	

	memset(filename, '\000', sizeof(filename));
	GetModuleFileName(mod.hDllInstance,filename,sizeof(filename));
	strcpy(currentDir, filename);
	char *pend;
	pend = strrchr(currentDir, '\\');
	if (pend) {
		*pend = '\000';
	}
	p = filename+lstrlen(filename);
	while (p >= filename && *p != '\\') p--;
	p++;

	char	logFile[1024] = "";
	char	configFile[1024] = "";
	char	currentlogFile[1024] = "";
	memset(logFile, '\000', sizeof(logFile));
	char *p2 = strchr(p, '.');
	if (p2) {
		strncpy(logFile, p, p2-p);
	}
	else {
		strcpy(logFile, p);
	}
	sprintf(configFile, "%s\\%s.cfg", currentDir, logFile);
	sprintf(currentlogFile, "%s\\%s.log", currentDir, logFile);

	setLogFile(currentlogFile);
	setConfigFileName(configFile);

	readConfigFile();

	ghwnd_winamp = this_mod->hwndParent;
	
//	wxEntry((WXHINSTANCE) mod.hDllInstance, (WXHINSTANCE) NULL, NULL, SW_SHOW, FALSE);

	oddcastHWND = CreateDialog(mod.hDllInstance,MAKEINTRESOURCE(IDD_ODDCAST), (HWND)this_mod->hwndParent, BigWindowHandler);
	SetWindowPos(oddcastHWND, this_mod->hwndParent, (int)getLastXWindow(), (int)getLastYWindow(), -1, -1, SWP_NOSIZE | SWP_SHOWWINDOW);

//	ShowWindow(oddcastHWND, SW_SHOW);

	initializeGlobals();
	transcode_init(0, NULL);

	if (getLockedMetadataFlag()) {
		setUIMetadata((char *)getLockedMetadata());
	}

	timerId = SetTimer(NULL, 1, 1000, (TIMERPROC)getCurrentSongTitle);
	reconnectTimerId = SetTimer(NULL, 2, getReconnectSecs()*1000, (TIMERPROC)ReconnectTimer);

	return 0;
}

// cleanup (opposite of init()). Destroys the window, unregisters the window class
void quitoddcast(struct winampDSPModule *this_mod)
{
	KillTimer(NULL, timerId);
	KillTimer(NULL, reconnectTimerId);
	RECT	pRect;
	if (oddcastHWND) {
		GetWindowRect(oddcastHWND, &pRect);

		setLastXWindow(pRect.left);
		setLastYWindow(pRect.top);
		DestroyWindow(oddcastHWND);
	}
	writeConfigFile();
	int a = 1;
	
}
static signed int scale(int sample)
{
	// round 
	sample += (1L << (MAD_F_FRACBITS - 16));

	// clip 
	if (sample >= MAD_F_ONE)
		sample = MAD_F_ONE - 1;
	else if (sample < -MAD_F_ONE)
		sample = -MAD_F_ONE;

	// quantize 
	return sample >> (MAD_F_FRACBITS + 1 - 16);
}
int encode_samples(struct winampDSPModule *this_mod, short int *short_samples, int numsamples, int bps, int nch, int srate)
{
	float	samples[8196*16];
	static short lastSample = 0;
	static signed int sample;

	int	samplecount = 0;
	short int *psample = short_samples;

	if (getIsConnected()) {
	  if (!getLiveRecordingFlag()) {
		  for (int i=0;i<numsamples*nch;i++) {
			sample = *psample++;
			samples[i] = sample/32767.f;
		  }
   		  int ret = handle_output((float *)&samples, numsamples, nch, srate);
	  }
  }
	return numsamples;
}	

void OnConnect()
{
		if (getIsConnected()) {
			stopOddcast();			
		}
		else {
			char	message[1024] = "";
			char	error[1024] = "";
			int ret = 1;
			if (getWMAFlag()) {
				ret = startWMAServer(message,error);
			}
			if (ret == 0) {
				MessageBox(oddcastHWND, message, "Message", MB_OK);
				MessageBox(oddcastHWND, error, "Message", MB_OK);
			}
			else {
				_beginthreadex(NULL, 0, (unsigned (_stdcall *) (void *))startOddcastThread, NULL, 0, &oddcastThread);
			}
		}

}
void OnConfig()
{
	char	commandToExecute[1024] = "";
	if (getIsConnected()) {
		MessageBox(oddcastHWND, "Cannot edit settings while connected", "Message", MB_OK);
	}
	else {

		memset(commandToExecute, '\000', sizeof(commandToExecute));
		sprintf(commandToExecute, "oddcast_editConfig.exe \"%s\"", getConfigFileName());

		STARTUPINFO si;
		PROCESS_INFORMATION pi;

		ZeroMemory( &si, sizeof(si) );
		si.cb = sizeof(si);
		ZeroMemory( &pi, sizeof(pi) );

		// Start the child process. 
		if( !CreateProcess( NULL, // No module name (use command line). 
			commandToExecute, // Command line. 
			NULL,             // Process handle not inheritable. 
			NULL,             // Thread handle not inheritable. 
			FALSE,            // Set handle inheritance to FALSE. 
			0,                // No creation flags. 
			NULL,             // Use parent's environment block. 
			NULL,             // Use parent's starting directory. 
			&si,              // Pointer to STARTUPINFO structure.
			&pi )             // Pointer to PROCESS_INFORMATION structure.
		) {
			MessageBox(oddcastHWND, "Cannot seem to find oddcast_editConfig.exe, please re-install oddcast.", "ERROR", MB_OK);
		}
		else {
				// Wait until child process exits.
				WaitForSingleObject( pi.hProcess, INFINITE );

				// Close process and thread handles. 
				CloseHandle( pi.hProcess );
				CloseHandle( pi.hThread );
				initializeGlobals();
				transcode_init(0, NULL);
				if (reconnectTimerId) {
					KillTimer(NULL, reconnectTimerId);
				}
				reconnectTimerId = SetTimer(NULL, 2, getReconnectSecs()*1000, (TIMERPROC)ReconnectTimer);

				/**********************************************************8
				FIX RECONNECT TIMER
				**********************************************************8
				**********************************************************8
				**********************************************************8
				**********************************************************8
				**********************************************************
				if (reconnectTimer.IsRunning()) {
					reconnectTimer.Stop();
				}
				if (getReconnectFlag()) {
					reconnectTimer.Start(getReconnectSecs()*1000, FALSE);
				}
				*/

		}
	}
}


void OnEditMetadata()
{
	DialogBox(mod.hDllInstance,MAKEINTRESOURCE(IDD_EDIT_METADATA), (HWND)oddcastHWND, EditMetadataHandler);
}
int CALLBACK EditMetadataHandler(HWND hwndDlg,UINT uMsg,WPARAM wParam,LPARAM lParam)
{
	int metadataLocked = 0;
	char	metadata[1024] = "";

	switch (uMsg) 
	{
		case WM_INITDIALOG:
			metadataLocked = getLockedMetadataFlag();
			if (metadataLocked) {
				CheckDlgButton(hwndDlg, IDC_LOCK_METADATA, BST_CHECKED);
				SetDlgItemText(hwndDlg, IDC_METADATA, (char *)getLockedMetadata());
			}
			else {
				CheckDlgButton(hwndDlg, IDC_LOCK_METADATA, BST_UNCHECKED);
				SetDlgItemText(hwndDlg, IDC_METADATA, (char *)getCurrentlyPlaying());
			}
			break;
		case WM_COMMAND:
			int	bleh = LOWORD(wParam);
			int bleh2 = HIWORD(wParam);
			switch (LOWORD(wParam))  
			{

			case IDOK:
				GetDlgItemText(hwndDlg, IDC_METADATA, metadata, sizeof(metadata)-1);
				setLockedMetadata((char *)metadata);
				setLockedMetadataFlag(IsDlgButtonChecked(hwndDlg, IDC_LOCK_METADATA));
				setUIMetadata((char *)metadata);
				writeConfigFile();
				EndDialog(hwndDlg,0);
				break;
			case IDCANCEL:
				EndDialog(hwndDlg,0);
				break;
			}
			break;
	}
	return FALSE;
}

int CALLBACK BigWindowHandler(HWND hwndDlg,UINT uMsg,WPARAM wParam,LPARAM lParam)
{
	char		tmp[10] = "";
	char	tempbuf[1024] = "";
	char	buffer[1024] = "";
	char	buf[25] = "";
	bool	ret = FALSE;
			char	bb[255] = "";
	int		rret = 0;
	char	dirpath[255] = "";
	char	ppath[1024] = "";
	short keydown = 0;
	char msg[255] = "";
	int	yesno = 0;

	switch (uMsg) 
	{

	case WM_INITDIALOG:
		break;
	case WM_COMMAND:
		
		int	bleh = LOWORD(wParam);
		int bleh2 = HIWORD(wParam);
		switch (LOWORD(wParam))  
		{

		case IDC_MANUALEDIT_METADATA:
			OnEditMetadata();
			break;
		case IDC_EDIT_CONFIG:
			OnConfig();
			break;
		case IDC_CONNECT:
			OnConnect();
			break;
		case IDOK:
			EndDialog(hwndDlg,0);
			break;

		case ID_QUIT:
			EndDialog(hwndDlg,0);
			break;
		}

	}
	return FALSE;
}
