#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <pthread.h>



#ifdef __cplusplus
extern "C" {
#endif
#include <xmms/plugin.h>
#include <xmmsctrl.h>
#include <xmms/configfile.h>
#include <gtk/gtk.h>
void oddcast_init();
void oddcast_configure();
void oddcast_cleanup();
void oddcast_about();
void oddcast_query_format(AFormat *fmt,gint *rate, gint *nch);
int oddcast_mod_samples(gpointer *data, gint numsamples, AFormat fmt, gint srate, gint nch);
int encode_samples(short int *samples, int numsamples, int bps, int nch, int srate);
void write_config();
void read_config();
int initializePipe();
FILE    *namedPipe = NULL;
FILE    *oddcastProcess = NULL;
char	gPipePath[1024] = "";
static 	int initialized = 0;
int		oddcast_pid = -1;
pthread_t 	id;

// Callback functions
EffectPlugin oddcastv2_vtable = {
	0, // Handle, filled in by xmms
	0, // Filename, filled in by xmms
	"Oddcast V2 Plugin",       // description
	oddcast_init,           // Called when plugin is enabled
	oddcast_cleanup,        // Called when plugin is disabled
	oddcast_about,          // Show the about box
	oddcast_configure,      // Show the configure box
	oddcast_mod_samples,     
	oddcast_query_format     
};

typedef struct OddAudiotag {
	int     nch;
	int     samplerate;
	int     numsamples;
	int     metadatalength;
	int     samplelength;
} OddAudio;
static GtkWidget *config_window = NULL;
static GtkWidget *pipe_path = NULL;
static GtkWidget *pipe_label = NULL;
static GtkWidget *save = NULL;
static GtkWidget *cancel = NULL;
static GtkWidget *fixed1 = NULL;

EffectPlugin *get_eplugin_info(void)
{
	return &oddcastv2_vtable;
}

#define APPNAME "oddcast_effect"


void on_save_clicked(GtkButton *button, gpointer user_data)
{
	memset(gPipePath, '\000', sizeof(gPipePath));
	strncpy(gPipePath, gtk_entry_get_text(GTK_ENTRY(pipe_path)), sizeof(gPipePath)-1);
	write_config();
	gtk_widget_destroy(config_window);
	config_window = NULL;
}
void on_cancel_clicked(GtkButton *button, gpointer user_data)
{
	gtk_widget_destroy(config_window);
	config_window = NULL;
}


void oddcast_display_config()
{

	if (config_window == NULL) {
		config_window = gtk_window_new (GTK_WINDOW_DIALOG);
		gtk_object_set_data (GTK_OBJECT (config_window), "config_window", config_window);
		gtk_window_set_title (GTK_WINDOW (config_window), "Configuration");

		fixed1 = gtk_fixed_new ();
		gtk_widget_ref (fixed1);
		gtk_object_set_data_full (GTK_OBJECT (config_window), "fixed1", fixed1, (GtkDestroyNotify) gtk_widget_unref);
		gtk_widget_show (fixed1);
		gtk_container_add (GTK_CONTAINER (config_window), fixed1);
		gtk_widget_set_usize (fixed1, 319, 104);


		pipe_label = gtk_label_new ("Communication Pipe - any changes here require a restart");
		gtk_widget_ref (pipe_label);
		gtk_object_set_data_full (GTK_OBJECT (config_window), "pipe_label", pipe_label, (GtkDestroyNotify) gtk_widget_unref);
		gtk_widget_show (pipe_label);
		gtk_fixed_put (GTK_FIXED (fixed1), pipe_label, 10, 10);
		gtk_widget_set_uposition (pipe_label, 10, 10);
		gtk_widget_set_usize (pipe_label, 288, 24);
		gtk_label_set_justify (GTK_LABEL (pipe_label), GTK_JUSTIFY_LEFT);
		gtk_label_set_line_wrap (GTK_LABEL (pipe_label), TRUE);

		pipe_path = gtk_entry_new ();
		gtk_widget_ref(pipe_path);
		gtk_object_set_data_full (GTK_OBJECT (config_window), "pipe_path", pipe_path, (GtkDestroyNotify) gtk_widget_unref);
		gtk_widget_show (pipe_path);
		gtk_fixed_put (GTK_FIXED (fixed1), pipe_path, 10, 30);
		gtk_widget_set_uposition (pipe_path, 10, 30);
		gtk_widget_set_usize (pipe_path, 240, 24);
		gtk_entry_set_text (GTK_ENTRY (pipe_path), gPipePath);

		save = gtk_button_new_with_label ("Ok");
		gtk_widget_ref (save);
		gtk_object_set_data_full (GTK_OBJECT (config_window), "save", save, (GtkDestroyNotify) gtk_widget_unref);
		gtk_widget_show (save);
		gtk_fixed_put (GTK_FIXED (fixed1), save, 10, 80);
		//gtk_fixed_put (GTK_FIXED (fixed2), save, 344, 8);
		gtk_widget_set_uposition (save, 10, 60);
		gtk_widget_set_usize (save, 72, 24);

		cancel = gtk_button_new_with_label ("Cancel");
		gtk_widget_ref (cancel);
		gtk_object_set_data_full (GTK_OBJECT (config_window), "cancel", cancel, (GtkDestroyNotify) gtk_widget_unref);
		gtk_widget_show (cancel);
		gtk_fixed_put (GTK_FIXED (fixed1), cancel, 50, 80);
		//gtk_fixed_put (GTK_FIXED (fixed2), cancel, 432, 8);
		gtk_widget_set_uposition (cancel, 100, 60);
		gtk_widget_set_usize (cancel, 72, 24);

		gtk_signal_connect (GTK_OBJECT (save), "clicked", GTK_SIGNAL_FUNC (on_save_clicked), NULL);
		gtk_signal_connect (GTK_OBJECT (cancel), "clicked", GTK_SIGNAL_FUNC (on_cancel_clicked), NULL);

		gtk_widget_show (config_window);
	}

}
void GetXMMSConfigVariable(char *varname, char *defaultvar, char *ptr, int len)
{
	ConfigFile *cfgfile;
	gchar   *value = NULL;

	if ((cfgfile = xmms_cfg_open_default_file()) != NULL) {
		xmms_cfg_read_string(cfgfile, APPNAME, varname, &value);
		if (!value) {
			strncpy(ptr, defaultvar, len);
		}
		else {
			strncpy(ptr, value, len);
		}
		g_free(value);
		xmms_cfg_free(cfgfile);
	}
}

void PutXMMSConfigVariable(char *varname, char *ptr)
{
	ConfigFile *cfgfile;

	if ((cfgfile = xmms_cfg_open_default_file()) != NULL) {
		xmms_cfg_write_string(cfgfile, APPNAME, varname, ptr);
		xmms_cfg_write_default_file(cfgfile);
		xmms_cfg_free(cfgfile);
	}
}


void read_config()
{
	memset(gPipePath, '\000' ,sizeof(gPipePath));

	GetXMMSConfigVariable("PipePath", "/tmp/oddcastFIFO", gPipePath, sizeof(gPipePath)-1);

}

void write_config()
{

	PutXMMSConfigVariable("PipePath", gPipePath);

}

char *getCurrentlyPlaying()
{
	gint pos,length;
	static gchar *title = NULL;
	static char	noInfo[25] = "No Info";

	if (title) {
		g_free(title);
		title = NULL;
	}
	if(xmms_remote_is_running(0)) {
		pos = xmms_remote_get_playlist_pos(0);
		title = xmms_remote_get_playlist_title(0,pos);
	}
	if (!title) {
		title = noInfo;
	}
	return title;
}


int terminatePipe()
{
	OddAudio	audioData;

	memset(&audioData, '\000', sizeof(audioData));
	audioData.nch = 0;
	audioData.samplerate = 0;
	audioData.numsamples = 0;
	audioData.samplelength = 0;

	audioData.metadatalength = strlen("ODDCAST_STOP_ME_NOW");

	fwrite(&audioData, 1, sizeof(audioData), namedPipe);
	if (audioData.metadatalength) { 
		fwrite("ODDCAST_STOP_ME_NOW", 1, audioData.metadatalength, namedPipe);
		fflush(namedPipe);
	}
	if (namedPipe) {
		fclose(namedPipe);
		namedPipe = NULL;
	}
	void	*returnVal;
	pthread_cancel(id);
	pthread_join(id, &returnVal);
	initialized = 0;
}
static void *
oddcastv2_thread (void *param)
{

	char command[1024] = "";
	char buf[1024] = "";
	FILE	*filep;
	system("mkdir ~/.oddcastv2 2>/dev/null");
	sprintf(command, "oddcastv2 %s ~/.oddcastv2/oddcast_xmms.log ~/.oddcastv2/oddcast_xmms.cfg", (char *)param);
	filep = popen(command, "r");
	while (!feof(filep)) {
		memset(buf, '\000', sizeof(buf));
		fgets(buf, sizeof(buf)-1, filep);
		fprintf(stdout, "%s\n", buf);
	}
	pthread_exit((void *)1);
}

int initializePipe()
{
	struct stat statbuf;
	char	buf[1024] = "";
	char	command[1024] = "";
	int ok = 1;

	if (strlen(gPipePath) == 0) {
		fprintf(stdout, "Unable to initialize, the pipe path is not-specified and is required!\n");
		return 0;
	}
	if (!initialized) {
		if (stat(gPipePath, &statbuf) == 0) {
			//remove(gPipePath);
			;
		}
		else {
		
				if (mkfifo(gPipePath, 0666) == 0) {
					;
				}
				else {
					sprintf(buf, "Unable to create FIFO for %s", gPipePath);
					ok = 0;
				}
		}

		namedPipe = fopen(gPipePath, "w+");
		if (namedPipe == 0) {
			sprintf(buf, "Error reading from pipe");
			ok = 0;
		}
		if (ok) {
				pthread_create (&id, NULL, oddcastv2_thread, gPipePath);
				initialized = 1;
		}
		else {
			fprintf(stdout, "%s\n", buf);
		}
	}
}
// Here is the entry point for the Plugin..this gets called first.
void oddcast_init(void)
{
	if (!initialized) {
		read_config();

		initializePipe();
	}
}

void oddcast_configure(void)
{
	oddcast_display_config();
}
// cleanup (opposite of init()). Destroys the window, unregisters the window class
void oddcast_cleanup()
{
	terminatePipe();	
}
void oddcast_about()
{
	;
}

int oddcast_mod_samples(gpointer *data, int length, AFormat fmt, gint srate, gint nch)
{
	OddAudio	audioData;
	static	char currentSong[1024] = "";

	memset(&audioData, '\000', sizeof(audioData));
	audioData.nch = nch;
	audioData.samplerate = srate;
	audioData.numsamples = length/(2*2);
	audioData.samplelength = length;
	float	*samples;
	short int *samples_int;


	if (!initialized) {
		initializePipe();
	}
	//encode_samples((short int *)*data, length/(2), 0, nch, srate);
	audioData.metadatalength = 0;
	if (strcmp(currentSong, getCurrentlyPlaying())) {
			memset(currentSong, '\000', sizeof(currentSong));
			strncpy(currentSong, getCurrentlyPlaying(), sizeof(currentSong)-1);
			audioData.metadatalength = strlen(currentSong);
	}

	audioData.numsamples = (length / sizeof(short))/nch;
	
	audioData.samplelength = nch*audioData.numsamples*(sizeof(float));
	samples = (float *)malloc(audioData.samplelength);
	samples_int = (short int *)*data;

	for (int i=0;i<audioData.numsamples*nch;i++) {
		samples[i] = samples_int[i]/32767.f;
	}

	if (namedPipe) {
			fwrite(&audioData, 1, sizeof(audioData), namedPipe);
			if (audioData.metadatalength) { 
					fwrite(currentSong, 1, audioData.metadatalength, namedPipe);
			}
			fwrite(samples, 1, audioData.samplelength, namedPipe);
	}
	if (samples) {
		free(samples);
	}
	return length;
}
void oddcast_query_format(AFormat *fmt, gint *rate, gint *nch)
{
	;
}
#ifdef __cplusplus
}
#endif
